/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2003-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.mathExpr;

import java.util.HashMap;

import de.jtem.mathExpr.evaluator.Context;
import de.jtem.mathExpr.evaluator.Evaluator;
import de.jtem.mathExpr.evaluator.FunctionCallEvaluator;
import de.jtem.mathExpr.evaluator.Type;
import de.jtem.mathExpr.evaluator.Variable;
import de.jtem.mathExpr.parser.Expression;
import de.jtem.mathExpr.parser.FunctionCall;
import de.jtem.mathExpr.parser.Symbol;

/**
 * Holds all defined {@link Variable Variables} and {@link Function Functions}.
 * With the methods {@link #defineVariable(String,Object,Type)} and
 * {@link #defineFunction(String,String[],Expression)} you can define new
 * <code>Variables</code> and <code>Functions</code>.
 * <p>
 * Usually you need a single instance of the class
 * {@link ExpressionConfiguration} which either automatically creates a new
 * instance of this class or gets a <code>DefinitionModel</code> by the
 * Constructor {@link ExpressionConfiguration#ExpressionConfiguration(
 * Type,DefinitionModel)}. Then you can use certain methods of the
 * <code>ExpressionConfiguration</code> class delegating appropriate methods 
 * in this class.
 * </p>
 * <p>
 * This class implements {@link Context} by resolving name symbols to
 * global <code>Variables</code> and function calls to <code>Functions</code>.
 * </p>
 *
 * @see ExpressionConfiguration#defineVariable(String,Object,Type)
 * @see ExpressionConfiguration#defineFunction(String,String[],String)
 * @see ExpressionConfiguration#defineFunction(String,String[],Expression)
 */
public class DefinitionModel implements Context
{
  /**
   * Stores all {@link Variable Variables} which are defined.
   */
  private HashMap varName2evaluator=new HashMap();
  /**
   * Stores all {@link Function Functions} which are defined.
   */
  private HashMap functionName2function=new HashMap();
  /**
   * Creates a new <code>DefinitionModel</code> which already holds all possible
   * {@link PredefinedFunction} Objects.
   * 
   * @see PredefinedFunction#FUNCTIONS
   */
  public DefinitionModel()
  {
    for(int i=0; i<PredefinedFunction.FUNCTIONS.length; i++)
      addFunction(PredefinedFunction.FUNCTIONS[i]);
  }
  /**
   * Defines and returns a new <code>Variable</code> with the specified 
   * <code>name</code>, <code>value</code> and <code>type</code>. If this 
   * <code>DefinitionModel</code> already contains a <code>Variable</code>
   * with the specified name, the <code>value</code> and <code>type</code>
   * of that <code>Variable</code> will be changed.
   *
   * @see Variable#Variable(String,Object,Type)
   * @see Variable#setValue(Object,Type)
   * @throws IncompatibleDefinitionException if a {@link Function} with the same
   * name is already defined.
   * @return the <code>Variable</code> which was defined respectively changed.
   * @param name the name for the <code>Variable</code> to be defined
   * respectively the name of the <code>Variable</code> to be changed.
   * @param value the value for the <code>Variable</code> to be defined
   * respectively the new value for the <code>Variable</code> to be changed.
   * @param type the <code>Type</code> for the <code>Variable</code> to be
   * defined respectively the new <code>Type</code> for the
   * <code>Variable</code> to be changed.
   */
  public Variable defineVariable(String name, Object value, Type type)
  {
    if(containsFunction(name))
      throw new IncompatibleDefinitionException(name, "Variable");
    //if(CAST_DEBUG)
    //  castLog("AbstractEvaluatorFactory.defineVariable");
    Variable v=getVariable(name);
    if(v==null)
      varName2evaluator.put(name, v=type.createVariable(name, value));
    else
    {
      v.setValue(value, type);
    }
    return v;
  }
  /**
   * Defines and returns a new <code>UserDefinedFunction</code> with the
   * specified <code>name</code>, <code>parameters</code> and
   * <code>definition</code>. If this <code>DefinitionModel</code> already
   * contains a <code>UserDefinedFunction</code> with the specified
   * <code>name</code>, the <code>parameters</code> and <code>definition</code>
   * of that <code>UserDefinedFunction</code> will be changed.
   *
   * @see UserDefinedFunction#UserDefinedFunction(String,String[],Expression)
   * @see UserDefinedFunction#setDefinition(Expression)
   * @see UserDefinedFunction#setParameters(String[])
   * @throws IllegalArgumentException if a {@link PredefinedFunction}
   * with the same name already exists.
   * @throws IncompatibleDefinitionException if a {@link Variable} with the same
   * name is already defined.
   * @return the <code>UserDefinedFunction</code> which was defined respectively
   * changed.
   * @param name the name for the <code>UserDefinedFunction</code> to be defined
   * respectively the name of the <code>UserDefinedFunction</code> to be
   * changed.
   * @param param an array of names of the parameters for the
   * <code>UserDefinedFunction</code> to be defined respectively the new
   * parameters for the <code>UserDefinedFunction</code> to be changed.
   * @param def the definition for the <code>UserDefinedFunction</code> to be
   * defined respectively the new definition for the <code>UserDefinedFunction
   * </code> to be changed.
   */
  public UserDefinedFunction defineFunction(String name, String[] param, 
  		Expression def)
  {
    if(containsVariable(name))
    {
      throw new IncompatibleDefinitionException(name, "Function");
    }
    //if(CAST_DEBUG)
    //  castLog("AbstractEvaluatorFactory.defineFunction");
    UserDefinedFunction f;
    try
    {
      f=(UserDefinedFunction)functionName2function.get(name);
    }
    catch(ClassCastException ex)
    {
      throw new IllegalArgumentException("PredefinedFunction "+
          name+" can't be redefined.");
    }
    if(f==null)
    {
      functionName2function.put(
          name, f=new UserDefinedFunction(name, param, def));
    }
    else
    {
      f.setDefinition(def);
      f.setParameters(param);
    }
    return f;
  }
  /**
   * Adds the specified <code>Variable</code> to this
   * <code>DefinitionModel</code>. If this <code>DefinitionModel</code> already
   * contains a <code>Variable</code> with the same name as
   * <code>newVariable</code>, value and type of that <code>Variable</code>
   * will be changed to value and type of <code>newVariable</code>.
   *
   * @see Variable#getValue()
   * @see Variable#getReturnType()
   * @see Variable#setValue(Object,Type)
   * @throws IncompatibleDefinitionException if a {@link Function} with the 
   * same name is already defined.
   * @param newVariable the <code>Variable</code> to be added to this
   * <code>DefinitionModel</code>
   */
  public void addVariable(Variable newVariable)
  {
    String name=newVariable.toString();
    if(containsFunction(name))
      throw new IncompatibleDefinitionException(name, "Variable");
    Variable v=getVariable(name);
    if(v==null)
      varName2evaluator.put(name, newVariable);
    else
    {
      v.setValue(newVariable.getValue(), newVariable.getReturnType());
    }
  }
  /**
   * Adds the specified <code>Function</code> to this
   * <code>DefinitionModel</code>. If this <code>DefinitionModel</code>
   * already contains a {@link UserDefinedFunction} with the same name as
   * <code>newFunction</code>, value and type of that
   * <code>UserDefinedFunction</code> will be changed to value and type of
   * <code>newFunction</code>.
   *
   * @see UserDefinedFunction#getDefinition()
   * @see UserDefinedFunction#getParameters()
   * @see UserDefinedFunction#setDefinition(Expression)
   * @see UserDefinedFunction#setParameters(String[])
   * @throws IllegalArgumentException if a {@link PredefinedFunction}
   * with the same name already exists.
   * @throws IncompatibleDefinitionException if a {@link Variable} with the 
   * same name is already defined.
   * @param newFunction the <code>Function</code> to be added to this
   * <code>DefinitionModel</code>
   */
  public void addFunction(Function newFunction)
  {
    String name=newFunction.getName();
    if(containsVariable(name))
    {
      throw new IncompatibleDefinitionException(name, "Function");
    }
    Function f=getFunction(name);
    if(f==null)
    {
      functionName2function.put(name, newFunction);
    }
    else
    {
      try
      {
        UserDefinedFunction uf=(UserDefinedFunction)f;
        uf.setDefinition(((UserDefinedFunction)newFunction).getDefinition());
        uf.setParameters(((UserDefinedFunction)newFunction).getParameters());
      }
      catch(ClassCastException ex)
      {
        throw new IllegalArgumentException("PredefinedFunction "+f+
            " already exists.");
      }
    }
  }
  /**
   * Returns the <code>Variable</code> with the specified <code>name</code> or
   * <code>null</code> if the <code>Variable</code> does not exist yet.
   * 
   * @return the <code>Variable</code> with the specified <code>name</code> or
   * <code>null</code> if the <code>Variable</code> does not exist yet.
   * @param name the name of the <code>Variable</code> this method tries to
   * return.
   */
  public Variable getVariable(String name)
  {
    //if(CAST_DEBUG)
    //  castLog("AbstractEvaluatorFactory.getVariable !!!");
    return (Variable)varName2evaluator.get(name);
  }
  /**
   * Returns the <code>Function</code> with the specified <code>name</code> or 
   * <code>null</code> if the <code>Function</code> does not exist yet.
   * 
   * @return the <code>Function</code> with the specified <code>name</code> or 
   * <code>null</code> if the <code>Function</code> does not exist yet.
   * @param name the name of the <code>Function</code> this method tries to
   * return.
   */
  public Function getFunction(String name)
  {
    //if(CAST_DEBUG)
    //  castLog("AbstractEvaluatorFactory.getFunction");
    Function function = (Function)functionName2function.get(name);
    //Pending need two methods
//    if(function==null)
//    	throw new NoSuchElementException("function "+name);
	return function;
  }
  /**
   * Returns <code>true</code> if this <code>DefinitionModel</code> contains a
   * {@link Variable} with the specified <code>name</code>; <code>false</code>
   * otherwise.
   * 
   * @see #defineVariable(String,Object,Type)
   * @see #addVariable(Variable)
   * @return <code>true</code> if this <code>DefinitionModel</code> contains a
   * <code>Variable</code> with the specified <code>name</code>;
   * <code>false</code> otherwise.
   * @param name the name of the <code>Variable</code> to be checked if
   * existing.
   */
  public boolean containsVariable(String name)
  {
    return varName2evaluator.containsKey(name);
  }
  /**
   * Returns <code>true</code> if this <code>DefinitionModel</code> contains a
   * {@link Function} with the specified <code>name</code>; <code>false</code>
   * otherwise.
   * 
   * @see #defineFunction(String,String[],Expression)
   * @see #addFunction(Function)
   * @return <code>true</code> if this <code>DefinitionModel</code> contains a
   * <code>Function</code> with the specified <code>name</code>;
   * <code>false</code> otherwise.
   * @param name the name of the <code>Function</code> to be checked if
   * existing.
   */
  public boolean containsFunction(String name)
  {
    return functionName2function.containsKey(name);
  }
  /**
   * Returns the {@link Variable} the specified <code>Symbol</code> is
   * representing.
   * <p>
   * This method is necessary for evaluating an {@link Expression} respectively
   * the <code>Symbols</code> in the expression-tree-hirarchy. This method
   * is called whenever a <code>Symbol</code> was found which represents a 
   * <code>Variable</code> in a global {@link Context}. This
   * <code>DefinitionModel</code> provides that <code>Context</code>.<br>
   * <br>
   * <i>Examples:</i><br>
   * 1. The <code>Expression</code> "2x+2y" contains the <code>Symbols</code>
   * "x" and "y" among other things. These <code>Symbols</code> represent
   * <code>Variables</code> in a global <code>Context</code> so that this method
   * will be called to get them.<br>
   * 2. A new {@link Function} "f(x,y)=2x+2y" was defined. Now the 
   * <code>Expression</code>  "f(1,1)" would implicate a {@link
   * FunctionCallEvaluator} which provides the <code>Context</code> for the
   * function's parameter "x" and "y". They do not represent global
   * <code>Variables</code>. If now "2x+2y" will be evaluated,  the 
   * <code>Evaluators</code> for "x" and "y" will be got by calling the method
   * {@link FunctionCallEvaluator#getEvaluator(Symbol)}.
   * </p>
   *
   * @throws UnknownDefinitionException if the <code>Variable</code> does not
   * exist in this <code>DefinitionModel</code>.
   * @return the <code>Variable</code> the specified <code>Symbol</code> is
   * representing.
   * @param s the <code>Symbol</code> representing the <code>Variable</code>
   * to be returned.
   */
  public Evaluator getEvaluator(Symbol s)
  {
    String str=s.toString();
    if(containsVariable(str))
      return getVariable(str);
    else
      throw new UnknownDefinitionException(str);
  }
  /**
   * Returns the <code>Function</code> according to the specified
   * <code>FunctionCall</code>. This method calls the method
   * {@link #getFunction(String)} parameterized with the name of the
   * <code>FunctionCall</code>.
   *
   * @see FunctionCallEvaluator#getFunction(FunctionCall)
   * @return the <code>Function</code> the specified
   * <code>FunctionCall</code> is representing.
   * @param fc the <code>FunctionCall</code> representing the
   * <code>Function</code> to return.
   */
  public Function getFunction(FunctionCall fc)
  {
    return getFunction(fc.getFunctionName());
  }
}
