/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2003-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.mathExpr;


import java.util.Arrays;

import de.jtem.mathExpr.evaluator.Context;
import de.jtem.mathExpr.evaluator.Evaluator;
import de.jtem.mathExpr.evaluator.FunctionCallEvaluator;
import de.jtem.mathExpr.evaluator.Type;
import de.jtem.mathExpr.parser.Expression;
//for doc-links:
import de.jtem.mathExpr.parser.FunctionCall;
import de.jtem.mathExpr.parser.List;
import de.jtem.mathExpr.parser.Symbol;

/**
 * A <code>UserDefinedFunction</code> object represents a mathematical function
 * with a name, one or more parameters and a definition (e.g.:
 * <code>f(x,y)=x+y</code>).
 */
public class UserDefinedFunction extends Function
{
  /**
   * The function's definition.
   */
  protected Expression definition;
  /**
   * Caches function derivatives created by getDerivative(int).
   */
  private UserDefinedFunction[] derivative;
  /**
   * Creates a new <code>UserDefinedFunction</code> with the specified
   * <code>name</code> and <code>parameters</code>. The definition
   * <code>Expression</code> remaines <code>null</code>.
   *
   * @see #setDefinition(Expression)
   * @param name the function's name.
   * @param param an array of the function's parameters.
   */
  public UserDefinedFunction(String name, String[] param)
  {
    super(name, param);
    derivative=new UserDefinedFunction[param.length];
  }
  /**
   * Creates a new <code>UserDefinedFunction</code> with the specified
   * <code>name</code>, <code>parameters</code> and <code>definition</code>.
   * 
   * @param name the function's name.
   * @param param an array of the function's parameters.
   * @param def the definition of this <code>UserDefinedFunction</code>.
   */
  public UserDefinedFunction(String name, String[] param, Expression def)
  {
    super(name, param);
    definition=def;
    derivative=new UserDefinedFunction[param.length];
  }
  /**
   * Sets the definition of this <code>UserDefinedFunction</code> to the
   * specified <code>Expression</code>.
   * 
   * @param newDef the new definition.
   */
  public void setDefinition(Expression newDef)
  {
    definition=newDef;
    Arrays.fill(derivative, null);
  }
  /**
   * Returns the Expression representing the function's definition.
   * 
   * @return the function's definition.
   */
  public Expression getDefinition()
  {
    return definition;
  }
  /**
   * Returns the <code>Evaluator</code> for this
   * <code>UserDefinedFunction</code> by calling
   * the method {@link Expression#getEvaluator(Type,Context)} on the 
   * {@link #definition} field. Usually the specified <code>Context</code>
   * will be the {@link FunctionCallEvaluator} for the {@link FunctionCall}
   * "calling" this <code>UserDefinedFunction</code>.
   *
   * @see FunctionCallEvaluator#getEvaluator(Symbol)
   * @see #getDefinition()
   * @return the <code>Evaluator</code> for the function's definition.
   * @param defaultType the <code>Type</code>, <code>Symbols</code> will be
   * evaluated to.
   * @param context the <code>FunctionCallEvaluator</code> for this
   * <code>UserDefinedFunction</code>.
   */
  public Evaluator getCallEvaluator(Type defaultType, Context context) 
  {
    return getDefinition().getEvaluator(defaultType, context);
  }
  /**
   * Returns the derivative function of this UserDefinedFunction with
   * respect to the paramIndex' parameter. The name of the returned 
   * function has the following syntax:
   * D&lt;paramIndex&gt;&lt;antiderivativeName&gt;
   * For example, if this Function has the name "f" and will be differentiated
   * with respect to its first parameter, the derivative will be "D0f".
   * 
   * @return a new Function which is the derivative of this with respect to the
   * paramIndex' parameter.
   */
  public Function getDerivative(int paramIndex)
  {
    if(derivative[paramIndex]==null)
    {
      Expression derivativeFunctionDefinition=definition.differentiate(
          new Symbol(parameter[paramIndex]));
      derivative[paramIndex]=new UserDefinedFunction(
          "D"+paramIndex+name, parameter, derivativeFunctionDefinition);
    }
    return derivative[paramIndex];
  }
  /**
   * Returns a dissolved FunctionCall of this Function with the specified
   * parameters. If param is an instance of List, the entries are the
   * parameters. Otherwise param ist only one parameter.
   */
  public Expression insertArguments(Expression param)
  {
    if(param instanceof List)
    {
      List parameters=(List)param;
      Expression dissolved=null;
      for(int i=0; i<parameters.getEntryCount(); i++)
      {
        Symbol p=new Symbol(parameter[i]);
        dissolved=dissolved==null ?
            definition.replaceSymbol(p, parameters.getEntry(i)) :
            dissolved.replaceSymbol(p, parameters.getEntry(i));
      }
      return dissolved;
    }
    //implementation for one parameter
    Symbol p=new Symbol(parameter[0]);
    return definition.replaceSymbol(p, param);
  }
  /**
   * Returns the <code>String</code> representing this
   * <code>UserDefinedFunction</code>.
   * Syntax: <code>name(parameter1[,parameter2,...])=definition</code>
   * 
   * @return the <code>String</code> representing this
   * <code>UserDefinedFunction</code>.
   */
  public String toString()
  {
    String str=getName()+"(";
    for(int i=0; i<parameter.length; i++)
    {
      str+=parameter[i];
      if(i<parameter.length-1)
        str+=",";
    }
    str+=")"+"="+getDefinition();
    return str;
  }
}
