/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2003-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.mathExpr.evaluator;

//for doc-links:
import de.jtem.mathExpr.ExpressionConfiguration;
import de.jtem.mathExpr.Function;
import de.jtem.mathExpr.parser.Expression;
import de.jtem.mathExpr.parser.Symbol;

/**
 * A <code>Variable</code> object represents a mathematical variable with a
 * name, value and a certain {@link Type}. The <code>Type</code> determines
 * what kind of <code>Object</code> is used for the variable's value. It's
 * also necessary for evaluating this <code>Variable</code> in relation to
 * other {@link Evaluator Evaluators} with different evaluation
 * <code>Types</code>.
 * <br><br>
 * Before evaluating an {@link Expression}, all {@link Symbol Symbols}, in
 * the expression-tree-hirarchy, which represent no constant numbers
 * respectively {@link Constant Constants} like "pi" or "e", must be defined
 * as a <code>Variable</code> object with the same name the appropriate
 * <code>Symbol</code> has.
 * 
 * @see Symbol#isConstant()
 * @see Symbol#toString()
 * @see #getName()
 * @see ExpressionConfiguration#defineVariable(String,Object,Type)
 * @see Function
 */
public class Variable implements Evaluator
{
	/**
	 * The value of this <code>Variable</code>.
	 */
  protected Object value;
  /**
   * The name of this <code>Variable</code>.
   */
  protected String name;
  /**
   * The <code>Type</code> of this <code>Variable</code>.
   */
  protected Type type;
  /**
   * Creates a new <code>Variable</code> with the name <code>name</code>
   * the value <code>value</code> and the <code>Type type</code>.
   * 
   * @param name the name for the new <code>Variable</code>.
   * @param value the value for the new <code>Variable</code>.
   * @param type the <code>Type</code> of the value of the new
   * <code>Variable</code>.
   */
  public Variable(String name, Object value, Type type)
  {
    this.value=value;
    this.name=name;
    this.type=type;
  }
  /**
   * Returns the value of this <code>Variable</code> as an <code>Object</code>.
   * The <code>Type</code> of that <code>Object</code> is returned by the
   * method {@link #getReturnType()}.
   * 
   * @return the value of this <code>Variable</code>.
   */
  public Object getValue()
  {
    return value;
  }
  /**
   * Sets the value of this <code>Variable</code> to the specified
   * <code>Object</code>. The specified <code>Type</code> is the
   * <code>Type</code> of <code>newVal</code>. Use this method,
   * if the <code>Type</code> of the new value differs from the old value
   * <code>Type</code>.
   *
   * @see #setValue(Object)
   * @param newVal the new value for this <code>Variable</code>.
   * @param newType the <code>Type</code> of the new value.
   */
  public void setValue(Object newVal, Type newType)
  {
    if(!value.equals(newVal))
      value=newVal;
    if(!type.equals(newType))
      type=newType;
  }
  /**
   * Sets the value of this <code>Variable</code> to the specified
   * <code>Object</code>. Use this method only if you are sure,
   * that the <code>Type</code> of the new value is equal to the old
   * value <code>Type</code>.
   *
   * @see #setValue(Object,Type)
   * @param newVal the new value for this <code>Variable</code>.
   */
  public void setValue(Object newVal)
  {
    if(!value.equals(newVal))
      value=newVal;
  } 
  /**
   * Returns the name of this <code>Variable</code>. The name of a any
   * <code>Variable</code> has to be defined by the constructor. A
   * <code>Variable</code> can't be renamed.
   * 
   * @return the name of this <code>Variable</code>.
   */
  public String getName()
  {
    return name;
  }
  /**
   * Evaluates this <code>Variable</code> by returning its value.
   * 
   * @return the value of this <code>Variable</code> as an <code>Object</code>.
   */
  public Object evaluate()
  {
    return getValue();
  }
  /**
   * Returns the <code>Type</code> of this <code>Variable's</code> value.
   * 
   * @return the <code>Type</code> of this <code>Variable's</code> value.
   */
  public Type getReturnType()
  {
    return type;
  }
  public boolean equals(Object o)
  {
    Variable v;
    try
    {
      v=(Variable)o;
    }
    catch(ClassCastException ex)
    {
      return false;
    }
    if(this.getName().equals(v.getName()) &&
        this.getValue().equals(v.getValue()) &&
        this.getReturnType().equals(v.getReturnType()))
      return true;
    else
      return false;
  }
  /**
   * Returns a <code>String</code> showing name and value of this
   * <code>Variable</code> in the following syntax: <code>name=value</code>.
   * 
   * @return a <code>String</code> showing name and value of this
   * <code>Variable</code>.
   */
  public String toString()
  {
    return getName()+"="+getValue();
  }
}
