/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2003-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.mathExpr.parser;


import de.jtem.mathExpr.Function;
import de.jtem.mathExpr.evaluator.Context;
import de.jtem.mathExpr.evaluator.Evaluator;
import de.jtem.mathExpr.evaluator.FunctionCallEvaluator;
import de.jtem.mathExpr.evaluator.Type;
//for doc-links:
import de.jtem.mathExpr.evaluator.realEvaluator.Real;
import de.jtem.mathExpr.parser.javacc.Parser;

/**
 * A <code>FunctionCall</code> is a <code>UnaryOperation</code> which
 * represents a mathematical function call (e.g.: f(2)). The example shows,
 * that an Expression, which looks like an implicit times operation, might be a
 * <code>FunctionCall</code> instead, if "f" is defined as a {@link Function}
 * by the user.
 * <br><br>
 * The operand of a <code>FunctionCall</code> is the <code>Expression</code>
 * representing the parameter values, e.g. f(2,3,4) - "f" is the function name,
 * "2,3,4" is the operand <code>Expression</code>.
 * <br><br>
 * Whenever the inherited method {@link UnaryOperation#getType} is called, it
 * will return the ASCI-code for 'f' as a representative for any
 * <code>FunctionCall</code>.
 *
 * @see ExpressionFactory#createFunctionCall(String)
 */
public class FunctionCall extends UnaryOperation
{
  /**
   * The <code>Function's</code> name.
   */
  String functionName;
  Function function;
  /**
   * Creates a new <code>FunctionCall</code> with the evaluation priority
   * <code>pri</code> and a <code>functionName</code>.
   * The operand of this <code>FunctionCall</code> will be
   * <code>null</code> until you call the method {@link
   * #setOperand(Expression)}. Use the other constructor, if
   * the operand is already known.
   * 
   * @param function The <code>Function</code> this <code>FunctionCall</code>
   * is calling.
   */
  public FunctionCall(Function function)
  {
    super('f', 8);
    functionName=function.getName();
    this.function=function;
  }
  /**
   * Creates a new <code>FunctionCall</code> with the evaluation priority
   * <code>pri</code>, a <code>functionName</code> and the operand
   * <code>paramExp</code>. Use the other constructor if the operand is not
   * known.
   * 
   * @param function The <code>Function</code> this <code>FunctionCall</code>
   * is calling.
   * @param paramExp the operand (parameter <code>Expression</code>) for this
   * <code>FunctionCall</code>.
   */
  public FunctionCall(Function function, List parameters)
  {
    super('f', 8, oneParam(parameters));
    functionName=function.getName();
    this.function=function;
  }
  //TODO functions with other than one parameter, Function must not be a subclass of unary operation
  private static Expression oneParam(List parameters) {
	if(parameters.getEntryCount()!=1)
		throw new UnsupportedOperationException("function with "+parameters.getEntryCount()+"parameters");
	return parameters.getEntry(0);
  }
/**
   * Returns the name of the {@link Function} this <code>FunctionCall</code>
   * "is calling".
   *
   * @see Function#getName()
   * @return the name of the {@link Function} this <code>FunctionCall</code>
   * "is calling".
   */
  public String getFunctionName()
  {
    return functionName;
  }
  /**
   * Returns the number of parameter values the operand {@link Expression}
   * of this <code>FunctionCall<code> contains.
   *
   * @see Function#getParameterCount()
   * @return the number of parameter values the operand <code>Expression</code>
   * of this <code>FunctionCall<code> contains.
   */
  public int getParameterCount()
  {
    if(operand instanceof List)
      return ((List)operand).getEntryCount();
    //TODO: what if e.g. "f()" (no parameters)? 
    return 1;
  }
  /**
   * Returns the <code>Evaluator</code> for this <code>FunctionCall</code>.
   * <code>Type defaultType</code> specifies, how to deal with
   * <code>Symbols</code> in the expression-tree-hirarchy (evaluate any
   * <code>Symbol</code> as a {@link de.jtem.mfc.field.Complex} or a {@link Real} or
   * whatever?).
   * 
   * @see FunctionCallEvaluator
   * @return the <code>Evaluator</code> for this <code>FunctionCall</code>.
   * @param defaultType the <code>Type</code> any <code>Symbol</code> will be
   * evaluated to.
   * @param context the <code>Context</code>, the <code>Evaluators<code> for
   * the <code>Symbols</code> are getting from.
   */
  public Evaluator getEvaluator(Type defaultType, Context context)
  {
    return defaultType.getEvaluator(this, context);
  }
  /**
   * Simplifies this FunctionCall's operand (the parameter expression) and
   * returns a new FunctionCall with it.
   */
  public Expression simplify()
  {
    return new FunctionCall(function, toList(operand.simplify()));
  }
  static List toList(Expression expression) {
	List l=new List();
	l.addEntry(expression);
	return l;
  }
/**
   * Returns the derivative of this function call. If the function
   * has more than one parameter, the derivative is the Addition
   * of the following Multiplications (x is the first parameter,
   * y the second one):<br>
   * f'(x,y)=f'*dx + f'*dy
   * 
   * @see Function#getDerivative(int)
   */
  public Expression differentiate(Symbol s)
  {
    Expression paramDerivative=operand.differentiate(s);

    if(operand instanceof List)
    {
      List paramDerivativeList=(List)paramDerivative;
      Function[] functionDerivatives=new Function[function.getParameterCount()];
      Multiplication[] chainRule=new Multiplication[functionDerivatives.length];
      for(int i=0; i<functionDerivatives.length; i++)
      {
        functionDerivatives[i]=function.getDerivative(i);
        Expression e=functionDerivatives[i].insertArguments(operand);
        chainRule[i]=new Multiplication(e, paramDerivativeList.getEntry(i));
      }
      Expression derivative=chainRule[0];
      
      for(int i=0; i<functionDerivatives.length-1; i++)
      {
        System.out.println("more than one parameter");
        derivative=new Addition(derivative, chainRule[i+1]);
      }
      return derivative.simplify();
    }
    //implementation for one parameter
    Function functionDerivative=function.getDerivative(0);
    Expression e=functionDerivative.insertArguments(operand);
    return new Multiplication(e, paramDerivative).simplify();
  }
  /**
   * Inserts all arguments of this FunctionCall into the function definition
   * and returns that new Expression. 
   * 
   * @see Function#insertArguments(Expression)
   */
  public Expression dissolveFunctionCalls()
  {
    return function.insertArguments(operand.dissolveFunctionCalls());
  }
  /**
   * Returns <code>true</code> if the specified <code>Object</code> is a
   * <code>FunctionCall</code> to the same funtion and with equal parameters,
   * <code>false</code> otherwise.
   * 
   * @param o the object to compare this <code>FunctionCall</code> against.
   * @return <code>true</code> if the specified <code>Object</code> is a
   * <code>FunctionCall</code> to the same funtion and with equal parameters,
   * <code>false</code> otherwise.
   */
  public boolean equals(Object o)
  {
    if(o==this) return true;
    if(!(o instanceof FunctionCall))
      return false;
    FunctionCall fc=(FunctionCall)o;
    if(!functionName.equals(fc.getFunctionName()))
      return false;
    if(!operand.equals(fc.getOperand()))
      return false;
    return true;
  }
  /**
   * Returns a <code>String</code> representing this
   * <code>FunctionCall</code>. It's now possible to parse the
   *  returned <code>String</code> by a {@link Parser} again.
   * 
   * @return a <code>String</code> representing this
   * <code>FunctionCall</code>.
   */
  public String toString()
  {
    char op=(char)type;
    String str;
    str="("+functionName+(operand!=null?(operand instanceof Symbol?
        "("+operand.toString()+")":operand.toString()):"?")+")";
    if(operand instanceof List)
    {
      int listOpenIndex=str.indexOf('{');
      int listCloseIndex=str.lastIndexOf('}');
      str=str.substring(0, listOpenIndex)+'('+
          str.substring(listOpenIndex+1, listCloseIndex)+')'+
          str.substring(listCloseIndex+1);
      //str=str.replace('{', '(');
      //str=str.replace('}', ')');
    }
    return str;
  }
}
