/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2003-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.mathExpr.parser;

/**
 * The BinaryOperation with the operator '-'. This class
 * implements the expression transformation methods 
 * {@link #simplify()} and {@link #differentiate(Symbol)} depending
 * on this certain Expression. 
 * 
 * @author marcel
 */
public class Subtraction extends BinaryOperation
{
  /**
   * Creates a new Subtraction object with no operands. See also constructor
   * {@link BinaryOperation#BinaryOperation(int, int)}.
   */
  public Subtraction()
  {
    super('-', 2);
  }
  /**
   * Creates a new Subtraction object with the specified operands. See also
   * constructor {@link
   * BinaryOperation#BinaryOperation(int, int, Expression, Expression)}.
   * 
   * @param op1
   * @param op2
   */
  public Subtraction(Expression op1, Expression op2)
  {
    super('-', 2, op1, op2);
  }
  /**
   * Returns a simplified Expression of this Subtraction. Therefor this
   * method calls simplify() on its operands. Then the following
   * possible Simplifications are:
   * <ul>
   * <li>a-0=a</li>
   * <li>0-a=-a</li>
   * <li>-a+b=b-a</li>
   * <li>-a-b=-(a+b)</li>
   * </ul>
   */
  public Expression simplify()
  {
    Expression newOperand1=operand1.simplify();
    Expression newOperand2=operand2.simplify();
    int minus=0;
    if(newOperand1 instanceof NegativeSign)
    {
      newOperand1=((UnaryOperation)newOperand1).getOperand();
      minus=minus | 1;
    }
    if(newOperand2 instanceof NegativeSign)
    {
      newOperand2=((UnaryOperation)newOperand2).getOperand();
      minus=minus | 2;
    }
      
    if(newOperand1 instanceof Symbol && newOperand1.isConstant())
    {
      Symbol s=(Symbol)newOperand1;
      if(s.getConstantValue()==0)
        return (minus&2)!=0 ? //minus==2 || minus==3 ?
            newOperand2 : new NegativeSign(newOperand2);
    }
    if(newOperand2 instanceof Symbol && newOperand2.isConstant())
    {
      Symbol s=(Symbol)newOperand2;
      if(s.getConstantValue()==0)
        return (minus&1)!=0 ? //minus==1 || minus==3 ?
            new NegativeSign(newOperand1) : newOperand1;
    }
    
    boolean equalOperands=newOperand1.equals(newOperand2);
    switch(minus)
    {
      case 0: 
        if(equalOperands) return new Symbol("0");
        else return new Subtraction(newOperand1, newOperand2);
      case 1: return new NegativeSign(new Addition(newOperand1, newOperand2));
      case 2: return new Addition(newOperand1, newOperand2);
      case 3:
        if(equalOperands) return new Symbol("0");
        else return new Subtraction(newOperand2, newOperand1);
      default: throw new IllegalStateException("never thrown exception");
    }
  }
  /**
   * Returns the derivative of this Subtraction with respect to s.
   * In other words: Returns a new Subtraction with this' operands
   * differentiated with respect to s.
   */
  public Expression differentiate(Symbol s)
  {
    return new Subtraction(
        operand1.differentiate(s), operand2.differentiate(s)).simplify();
  }
}
