/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2003-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.mathExpr.parser;

import java.util.Collection;

import de.jtem.mathExpr.evaluator.Context;
import de.jtem.mathExpr.evaluator.Evaluator;
import de.jtem.mathExpr.evaluator.Type;
//for doc-links:
import de.jtem.mathExpr.evaluator.Variable;
import de.jtem.mathExpr.evaluator.realEvaluator.Real;
import de.jtem.mathExpr.parser.javacc.Parser;

/**
 * A <code>UnaryOperation</code> is an <code>Expression</code> which
 * represents a mathematical unary operation (positive sign, negative sign or
 * a <code>FunctionCall</code>) of another <code>Expression</code>
 * (the operand).
 *
 * @see ExpressionFactory#createUnaryOperation(int)
 */
public abstract class UnaryOperation extends Expression
{
  /**
   * This <code>UnaryOperation's</code> operand.
   */
  protected Expression operand;
  /**
   * The operation type of this <code>UnaryOperation</code>.
   * 
   * @see #getType()
   */
  protected int type;//TODO: add appropriate final int values in this class, marcel!
  /**
   * The evaluation priority of this UnaryOperation.
   */
  private int priority;
  /**
   * Creates a new <code>UnaryOperation</code> with the operation type
   * <code>type</code> and the evaluation priority <code>pri</code>.
   * The operand of this <code>UnaryOperation</code> will be
   * <code>null</code> until you call the method {@link
   * #setOperand(Expression)}. Use the other constructor, if
   * the operand is already known.
   * 
   * @param type the operation type this <code>UnaryOperation</code> will
   * represent.
   * @param pri the evaluation priority this <code>UnaryOperation</code> will
   * have.
   */
  public UnaryOperation(int type, int pri)
  {
    this.type=type;
    priority=pri;
  }
  /**
   * Creates a new <code>UnaryOperation</code> with the operation type
   * <code>type</code> , the evaluation priority <code>pri</code> and the
   * operand <code>op</code>. Use the other constructor if the operand is
   * not known.
   * 
   * @param type the operation type this <code>UnaryOperation</code> will
   * represent.
   * @param pri the evaluation priority this <code>UnaryOperation</code> will
   * have.
   * @param op the operand for this <code>UnaryOperation</code>.
   */
  public UnaryOperation(int type, int pri, Expression op)
  {
    this(type, pri);
    operand=op;
  }
  /**
   * Sets the operand of this <code>UnaryOperation</code> to the specified
   * <code>Expression</code>.
   * 
   * @param op the new operand for this <code>UnaryOperation</code>.
   */
  public void setOperand(Expression op)
  {
    operand=op;
  }
  /**
   * Returns the operand of this <code>UnaryOperation</code> or
   * <code>null</code> if there is no operand defined yet. Define the
   * operand with the appropriate constructor or with the method
   * {@link #setOperand(Expression)}.
   * 
   * @return the operand of this <code>UnaryOperation</code>.
   */
  public Expression getOperand()
  {
    return operand;
  }
  /**
   * Returns the operation type of this <code>UnaryOperation</code>.
   * It's the ASCII-code of the following characters representing the
   * appropriate operation:<br><br>
   * &nbsp;&nbsp;&nbsp;&nbsp;'+': positiv algebraic sign<br>
   * &nbsp;&nbsp;&nbsp;&nbsp;'-': negativ algebraic sign
   * 
   * @return the operation type of this <code>UnaryOperation</code>.
   */
  public int getType()
  {
    return type;
  }
  /**
   * Returns the evaluation priority of this <code>UnaryOperation</code>.
   * 
   * @return the evaluation priority of this <code>UnaryOperation</code>.
   */
  public int getPriority()
  {
    return priority;
  }
  /**
   * Returns <code>true</code> if the operand of this
   * <code>UnaryOperation</code> is not null <b>and</b> also constant,
   * <code>false</code> if not.
   * 
   * @return <code>true</code> if the operand of this
   * <code>UnaryOperation</code> is not null <b>and</b> also constant,
   * <code>false</code> if not.
   */
  public boolean isConstant()
  {
    return operand!=null && operand.isConstant();
  }
  /**
   * Returns <code>true</code> if the operand of this
   * <code>UnaryOperation</code> is not null <b>and</b> also complete,
   * <code>false</code> otherwise.
   * 
   * @return <code>true</code> if the operand of this
   * <code>UnaryOperation</code> is not null <b>and</b> also complete,
   * <code>false</code> otherwise.
   */
  public boolean isComplete()
  {
    return operand!=null && operand.isComplete();
  }
  /**
   * Adds the names of the {@link Variable Variables} which are essential for
   * evaluation into the specified <code>Collection</code>. Therefor it gets the
   * essential <code>Variables</code> of the operand of this
   * <code>UnaryOperation</code>.
   * 
   * @param c the <code>Collection</code> the <code>Variable</code> names
   * will be added to.
   */
  public void getEssentialVariables(Collection c)
  {
    operand.getEssentialVariables(c);
  }
  /**
   * Returns the <code>Evaluator</code> for this <code>UnaryOperation</code>.
   * <code>Type defaultType</code> specifies, how to deal with
   * <code>Symbols</code> in the expression-tree-hirarchy (evaluate any
   * <code>Symbol</code> as a {@link de.jtem.mfc.field.Complex} or a {@link Real} or
   * whatever?). It has nothing to do with {@link #getType()}.
   * 
   * @return the <code>Evaluator</code> for this <code>UnaryOperation</code>.
   * @param defaultType the <code>Type</code> any <code>Symbol</code> will be
   * evaluated to.
   * @param context the <code>Context</code>, the <code>Evaluators<code> for
   * the <code>Symbols</code> are getting from.
   */
  public Evaluator getEvaluator(Type defaultType, Context context)
  {
    Evaluator e=operand.getEvaluator(defaultType, context);
    return e.getReturnType().getEvaluator(this, e);
  }
  /**
   * Replaces all Symbols equal to s in this UnaryOperation
   * with the specified Expression. Therefor it replaces
   * all Symbols equal to s in this' operand. 
   * 
   * @see Symbol#replaceSymbol(Symbol,Expression)
   */
  public Expression replaceSymbol(Symbol s, Expression expr)
  {
    UnaryOperation newUO=null;
    try {
      newUO=(UnaryOperation)(this.clone()); }
    catch(CloneNotSupportedException ex){}
    newUO.setOperand(operand.replaceSymbol(s, expr));
    return newUO;
  }
  
  public Expression dissolveFunctionCalls()
  {
    UnaryOperation newUO=null;
    try {
      newUO=(UnaryOperation)(this.clone()); }
    catch(CloneNotSupportedException ex){}
    Expression newOp=operand.dissolveFunctionCalls();
    newUO.setOperand(newOp);
    return newUO;
  }
  /**
   * Returns <code>true</code> if the specified <code>Object</code> is a
   * <code>UnaryOperation</code> of the same operation type and with an equal
   * operand, <code>false</code> otherwise.
   * 
   * @param o the object to compare this <code>UnaryOperation</code> against.
   * @return <code>true</code> if the specified <code>Object</code> is a
   * <code>UnaryOperation</code> of the same operation type and with an equal
   * operand, <code>false</code> otherwise.
   */
  public boolean equals(Object o)
  {
    if(o==this) return true;
    if(!(o instanceof UnaryOperation))
      return false;
    UnaryOperation uo=(UnaryOperation)o;
    if(type!=uo.getType())
      return false;
    if(!operand.equals(uo.getOperand()))
      return false;
    return true;
  }
  /**
   * Returns a <code>String</code> representing this
   * <code>UnaryOperation</code>. It's now possible to parse the
   *  returned <code>String</code> by a {@link Parser} again.
   * 
   * @return a <code>String</code> representing this
   * <code>UnaryOperation</code>.
   */
  public String toString()
  {
    char op=(char)type;
    String str;
      str="("+op+' '+(operand!=null? operand.toString(): "?")+")";
    return str;
  }
}
